from operator import index
import queue
import numpy as np
import copy
import logging
import networkx as nx
from aspmc.util import *


logger = logging.getLogger("aspmc")

class Node(object):
    """A node class corresponding to nodes in a `Circuit`.
    
    Args:
        type (:obj:`int`): Which type the node has. One of `Node.AND, Node.OR, Node.NEG, Node.IN`.
        name (:obj:`int`): Unique identifier of the node.
        children (:obj:`iterable`, optional): The children of the node. Defaults to `[]`.
        ancestors (:obj:`iterable`, optional): The ancestors of the node. Defaults to `[]`.

    Attributes:
        type (:obj:`int`): Which type the node has. One of `Node.AND, Node.OR, Node.NEG, Node.IN`.
        name (:obj:`int`): Unique identifier of the node. 
        If `type` is `Node.NEG` or `Node.IN` this is the name of the (negated) input variable.
        children (:obj:`iterable`): The children of the node.
        ancestors (:obj:`iterable`): The ancestors of the node. 

    """
    AND = 0
    """Node type `AND` with value 0. Means that the children are conjoined."""
    OR = 1
    """Node type `OR` with value 1. Means that the children are disjoined."""
    NEG = 2
    """Node type `NEG` with value 2. Means that this is a negative input variable."""
    IN = 3
    """Node type `IN` with value 3. Means that this is a positive input variable."""
    def __init__(self, type, name, children = [], ancestors = []):
        self.type = type
        self.name = name
        self.children = children
        self.ancestors = ancestors

class Circuit(object):    
    """A circuit class that can store for example d-DNNFs and SDDs. 

    Offers methods to perform algebraic model counting on the circuit in linear time in the size of the (smooth) circuit.
    Also offers the same methods statically to avoid storing the circuit in memory. 
    Then the algebraic model counting is performed while reading the circuit.

    Supports the d-DNNF format of c2d and miniC2D for constructing circuits.
    
    Args:
        path (:obj:`string`): The path to the file that contains the circuit.

    Attributes:
        literals (:obj:`list`): The literals in the circuit. The node for literal `v` is in `literals[2*(v-1)]`, the one for `-v` is in `literals[2*(v-1)+1]`
        nodes (:obj:`list`): All the nodes of the circuit that are not literals.
    """
    def __init__(self, path):
        with open(path) as ddnnf:
            _, _, _, nr_leafs = ddnnf.readline().split()
            self.literals = []
            for var in range(1, int(nr_leafs) + 1):
                self.literals.append(Node(Node.IN, var))
                self.literals.append(Node(Node.NEG, -var))
            self.nodes = []
            idx = 0
            for line in ddnnf:
                line = line.strip().split()
                if line[0] == 'L':
                    node = self.literals[to_pos(int(line[1]))]
                    node.vars = (abs(int(line[1])),)
                elif line[0] == 'A':
                    node = Node(Node.AND, idx, children = [self.nodes[int(x)] for x in line[2:]])
                    node.vars = set()
                    for child in node.children:
                        child.ancestors.append(node)
                        node.vars.update(child.vars)
                    node.vars = tuple(node.vars)
                elif line[0] == 'O':
                    node = Node(Node.OR, idx, children = [self.nodes[int(x)] for x in line[3:]])
                    node.vars = set()
                    for child in node.children:
                        child.ancestors.append(node)
                        node.vars.update(child.vars)
                    node.vars = tuple(node.vars)
                self.nodes.append(node)
                idx += 1

    def non_smooth_wmc(self, weights):    
        """Performs weighted model counting over a circuit that is not smooth. Should be avoided since it is quite inefficient.
        
        Args:
            weights (:obj:`string`): The weights of the literals. 
                The weight for literal `v` is in `weights[2*(v-1)]`, the one for `-v` is in `weights[2*(v-1)+1]`

        Returns:
            (:obj:`object`): The weighted model count.
        """
        for i in range(len(self.literals)//2):
            self.literals[i*2].weight = weights[i*2]
            self.literals[i*2 + 1].weight = weights[i*2 + 1]

        todo = [ len(node.children) for node in self.nodes ]
        q = queue.Queue()
        for i in range(len(self.nodes)):
            if todo[i] == 0:
                q.put(self.nodes[i])
        while not q.empty():
            node = q.get()
            for anc in node.ancestors:
                todo[anc.name] -= 1
                if todo[anc.name] == 0:
                    q.put(anc)
            if node.type == Node.AND:
                value = np.full(len(weights[0]), 1.0)
                for child in node.children:
                    value *= child.weight
                node.weight = value
            elif node.type == Node.OR:
                value = np.full(len(weights[0]), 0.0)
                for child in node.children:
                    to_add = copy.deepcopy(child.weight)
                    for i in set(node.vars).difference(set(child.vars)):
                        to_add *= self.literals[to_pos(i)].weight + self.literals[neg(to_pos(i))].weight
                    value += to_add
                node.weight = value
        for i in set(range(1, 1 + len(self.literals)//2)).difference(set(node.vars)):
            node.weight *= self.literals[to_pos(i)].weight + self.literals[neg(to_pos(i))].weight
        return node.weight

    def wmc(self, weights):
        """Performs algebraic model counting over a circuit that is smooth.
        
        Args:
            weights (:obj:`string`): The weights of the literals. The weight for literal `v` is in `weights[2*(v-1)]`, the one for `-v` is in `weights[2*(v-1)+1]`

        Returns:
            (:obj:`object`): The algebraic model count.
        """
        for i in range(len(self.literals)//2):
            self.literals[i*2].weight = weights[i*2]
            self.literals[i*2 + 1].weight = weights[i*2 + 1]

        todo = [ len(node.children) for node in self.nodes ]
        q = queue.Queue()
        for i in range(len(self.nodes)):
            if todo[i] == 0:
                q.put(self.nodes[i])
        while not q.empty():
            node = q.get()
            for anc in node.ancestors:
                todo[anc.name] -= 1
                if todo[anc.name] == 0:
                    q.put(anc)
            if node.type == Node.AND:
                node.weight = np.full(len(weights[0]), 1.0)
                for child in node.children:
                    node.weight *= child.weight
            elif node.type == Node.OR:
                node.weight = np.full(len(weights[0]), 0.0)
                for child in node.children:
                    node.weight += child.weight
        return node.weight

    @staticmethod
    def parse_wmc(path, weights, zero = 0.0, one = 1.0, dtype = float, solver="c2d", vtree = None):
        """Performs algebraic model counting over a circuit that is smooth while parsing it.
        
        Args:
            path (:obj:`string`): The path to the file that contains the circuit.
            weights (:obj:`string`): The weights of the literals. The weight for literal `v` is in `weights[2*(v-1)]`, the one for `-v` is in `weights[2*(v-1)+1]`
            zero (:obj:`object`, optional): The neutral element of addition. Defaults to `0.0`.
            one (:obj:`object`, optional): The neutral element of multiplication. Defaults to `1.0`.
            dtype (:obj:`type`, optional): Which type the numpy arrays used to store the weights should have. Defaults to `float`.
            solver (:obj:`string`, optional): Which knowledge compiler this d-DNNF is from. Defaults to `c2d`.
            vtree (:obj:`aspmc.compile.vtree.Vtree`, optional): The underlying vtree if this is a non-smooth SDD. Defaults to `None`.
        Returns:
            (:obj:`object`): The algebraic model count.
        """
        if solver == "d4":
            return Circuit._parse_wmc_d4(path, weights, zero = zero, one = one, dtype = dtype)
        if solver == "miniC2D":
            return Circuit._parse_wmc_miniC2D(path, weights, zero = zero, one = one, dtype = dtype, vtree = vtree)
        shape = (np.shape(weights[0])[0], ) + np.shape(one)
        with open(path) as ddnnf:
            _, v, e, n = ddnnf.readline().split()
            logger.debug(f"d-DNNF size: {v} nodes, {e} edges, {n} variables")
            mem = [] # np.empty((int(v), ) + shape, dtype=dtype) (this seems to perform worse than using mem.append() in a loop !?)
            idx = 0
            for line in ddnnf:
                line = line.strip().split()
                mem.append(np.empty(shape, dtype = dtype))
                if line[0] == 'L':
                    mem[idx][:] = weights[to_pos(int(line[1]))]
                elif line[0] == 'A':
                    mem[idx][:] = one
                    for x in line[2:]:
                        mem[idx] *= mem[int(x)]
                elif line[0] == 'O':
                    mem[idx][:] = zero
                    for x in line[3:]:
                        mem[idx] += mem[int(x)]
                idx += 1
            return mem[idx - 1]

    @staticmethod
    def live_parse_wmc(pipe, weights, zero = 0.0, one = 1.0, dtype = float):
        """Performs algebraic model counting over a circuit that is smooth while parsing it.
        
        Args:
            pipe (:obj:`stream`): The pipe where the circuit is printed.
            weights (:obj:`string`): The weights of the literals. The weight for literal `v` is in `weights[2*(v-1)]`, the one for `-v` is in `weights[2*(v-1)+1]`
            zero (:obj:`object`, optional): The neutral element of addition. Defaults to `0.0`.
            one (:obj:`object`, optional): The neutral element of multiplication. Defaults to `1.0`.
            dtype (:obj:`type`, optional): Which type the numpy arrays used to store the weights should have. Defaults to `float`.
        Returns:
            (:obj:`object`): The algebraic model count.
        """
        mem = [] 
        shape = (np.shape(weights[0])[0], ) + np.shape(one)
        idx = 0
        for line in pipe:
            line = line.decode().strip().split()
            if line[0] == 'c':
                continue
            mem.append(np.empty(shape, dtype = dtype))
            if line[0] == 'L':
                mem[idx][:] = weights[to_pos(int(line[1]))]
            elif line[0] == 'A':
                mem[idx][:] = one
                for x in line[2:]:
                    mem[idx] *= mem[int(x)]
            elif line[0] == 'O':
                mem[idx][:] = zero
                for x in line[3:]:
                    mem[idx] += mem[int(x)]
            idx += 1
        return mem[idx - 1]

    @staticmethod
    def _parse_wmc_d4(path, weights, zero = 0.0, one = 1.0, dtype = float):
        shape = (np.shape(weights[0])[0], ) + np.shape(one)
        with open(path) as ddnnf:
            mem = [ None ] # the values of the nodes
            mem_type = [ False ] # whether the nodes are multiplicative
            for line in ddnnf:
                line = line.strip().split()
                if line[0] == 't': # add a true node
                    mem.append(np.full(shape, one, dtype=dtype))
                    mem_type.append(True)
                elif line[0] == 'f': # add a false node
                    mem.append(np.full(shape, zero, dtype=dtype))
                    mem_type.append(False)
                elif line[0] == 'a': # add an and node
                    mem.append(np.full(shape, one, dtype=dtype))
                    mem_type.append(True)
                elif line[0] == 'o': # add an or node
                    mem.append(np.full(shape, zero, dtype=dtype))
                    mem_type.append(False)
                elif mem_type[int(line[0])]: # add edge of an and node
                    mem[int(line[0])] *= mem[int(line[1])]
                else: # add edge of an or node                    
                    val = copy.deepcopy(mem[int(line[1])])
                    idx = 2
                    lit = int(line[idx])
                    while lit != 0:
                        val *= weights[to_pos(lit)]
                        idx += 1
                        lit = int(line[idx])
                    idx += 1
                    var = int(line[idx])
                    while var != 0:
                        val *= weights[to_pos(var)] + weights[neg(to_pos(var))]
                        idx += 1
                        var = int(line[idx])
                    mem[int(line[0])] += val
            return mem[1]

    @staticmethod
    def _parse_wmc_miniC2D(path, weights, zero = 0.0, one = 1.0, dtype = float, vtree = None):
        # setup phase: 
        # get the least common ancestors to speed up queries
        vtree_graph = vtree.to_networkx()
        vtree.index()
        lca = [ [ -1 for _ in range(vtree.leaf_count()*2) ] for _ in range(vtree.leaf_count()*2) ]
        for p,l in nx.tree_all_pairs_lowest_common_ancestor(vtree_graph):
            lca[p[1]][p[0]] = l
            lca[p[0]][p[1]] = l

        shape = (np.shape(weights[0])[0], ) + np.shape(one)

        index_to_node = { node.idx : node for node in vtree }
        index_to_node = [ index_to_node[i] for i in range(1, vtree.leaf_count()*2) ]
        # method to get the weight of the free variables not included in the SDD between node i and j
        def factor(i, j):
            res = np.full(shape, one, dtype=dtype)
            if lca[i][j] == j: # i should always be above j
                i,j = j,i
            if i == j:
                return res
            upper = index_to_node[i - 1]
            lower = index_to_node[j - 1]
            stack = queue.LifoQueue()
            cur = upper
            down = True
            while cur != upper or down:
                if down:
                    if cur == lower:
                        down = False
                    elif cur.val is None:
                        stack.put(cur)
                        cur = cur.left
                    else:
                        res *= weights[to_pos(cur.val)] + weights[neg(to_pos(cur.val))]
                        down = False
                else:
                    last = cur
                    cur = stack.get()
                    if cur.left == last:
                        stack.put(cur)
                        cur = cur.right
                        down = True
            return res

        # the actual counting phase
        with open(path) as ddnnf:
            _, v, e, n = ddnnf.readline().split()
            logger.debug(f"d-DNNF size: {v} nodes, {e} edges, {n} variables")
            mem = []
            vtree_nodes = []
            idx = 0
            for line in ddnnf:
                line = line.strip().split()
                if line[0] == 'L':
                    val = weights[to_pos(int(line[1]))]
                    vtree_node = abs(int(line[1]))
                else:
                    if line[0] == 'A':
                        vtree_node = lca[vtree_nodes[int(line[2])]][vtree_nodes[int(line[3])]]
                        left = index_to_node[vtree_node-1].left.idx
                        right = index_to_node[vtree_node-1].right.idx
                        if lca[right][vtree_nodes[int(line[2])]] != right:
                            right, left = left, right
                        val = mem[int(line[2])]*factor(right, vtree_nodes[int(line[2])])*mem[int(line[3])]*factor(left, vtree_nodes[int(line[3])])
                    elif line[0] == 'O':
                        vtree_node = lca[vtree_nodes[int(line[3])]][vtree_nodes[int(line[4])]]  
                        for x in line[3:]:
                            vtree_node = lca[vtree_node][vtree_nodes[int(x)]]         
                        val = np.full(shape, zero, dtype=dtype)
                        for x in line[3:]:
                            val += mem[int(x)]*factor(vtree_node, vtree_nodes[int(x)])
                mem.append(val)
                vtree_nodes.append(vtree_node)
                idx += 1
            mem[idx - 1] *= factor(vtree_node, vtree.idx)
            return mem[idx - 1]



from dataclasses  import dataclass
from collections import namedtuple
from typing import List, Iterable, Union, Dict
from clingo import Symbol, TruthValue, Function # pylint: disable=import-error, unused-import, no-name-in-module
from aspmc.parsing.clingoparser.literals import Literal

class ClingoObject(object):
    pass


@dataclass
class ClingoOutputAtom(ClingoObject):
    symbol: Symbol
    atom: int
    order: int = 0

    def __lt__(self, other):
        if self.__class__ == other.__class__:
            return (self.symbol, self.atom) < (other.symbol, other.atom)
        elif isinstance(other, ClingoObject):
            return self.order < other.order
        raise Exception("Incomparable type")


class ClingoRuleABC(ClingoObject):
    pass


@dataclass
class ClingoRule(ClingoRuleABC):
    choice: bool
    head: List[int]
    body: List[int]
    order: int = 1
    
    def __lt__(self, other):
        if self.__class__ == other.__class__:
            return (self.choice, self.head, self.body) < (other.choice, other.head, other.body)
        elif isinstance(other, ClingoObject):
            return self.order < other.order
        raise Exception("Incomparable type")


@dataclass
class ClingoProject(ClingoObject):
    atoms: List[int]
    order: int = 3

    def __lt__(self, other):
        if self.__class__ == other.__class__:
            return self.atoms < other.atoms
        elif isinstance(other, ClingoObject):
            return self.order < other.order
        raise Exception("Incomparable type")


ClingoExternal   = namedtuple('ClingoExternal', ['atom', 'value'])
ClingoHeuristic  = namedtuple('ClingoHeuristic', ['atom', 'type', 'bias', 'priority', 'condition'])
ClingoMinimize   = namedtuple('ClingoMinimize', ['priority', 'literals'])
ClingoWeightRule = namedtuple('ClingoWeightRule', ['choice', 'head', 'body', 'lower'])
ClingoAssume     = namedtuple('ClingoAssume', ['literals'])

@dataclass
class GroundProgram():
    objects : List[ClingoObject]

    def __init__(self, objects: Iterable[ClingoObject] = []):  # pylint: disable=dangerous-default-value
        self.objects = list(objects)

    def add_rule(self, choice: bool = False, head: Iterable[int] = [], body: Iterable[int] = []) -> None: # pylint: disable=dangerous-default-value
        self.objects.append(ClingoRule(choice=choice, head=list(head), body=list(body)))

    def add_rules(self, rules: Iterable[ClingoRule]) -> None:
        self.objects.extend(rules)

    def add_project(self, atoms: List[int] = []) -> None: # pylint: disable=dangerous-default-value
        self.objects.append(ClingoProject(list(atoms)))


    def add(self, obj: Union[ClingoObject, Iterable[ClingoObject]]) -> None:
        if isinstance(obj, ClingoObject):
            self.objects.append(obj)
        elif isinstance(obj, Iterable): # pylint: disable=isinstance-second-argument-not-valid-type
            self.objects.extend(obj)

    def __str__(self):
        return str(PrettyGroundProgram(self.objects))

    def __iter__(self):
        return iter(self.objects)

    # def add_external(self, external: Union[External, ClingoExternal]) -> None:
    #     if isinstance(external, ClingoExternal):
    #         external = self._rewrite_external(external)
    #     self.externals.append(external)

    # def add_assume(self, literals: List[int]) -> None:
    #     self.assumtions.append(ClingoAssume(literals))

    # def add_external(self, atom: int, value: TruthValue = TruthValue.False_) -> None:
    #     self.externals.append(ClingoExternal(atom, value))

    # def add_heuristic(self, atom: int, type: HeuristicType, bias: int, priority: int, condition: List[int]) -> None:
    #     self.externals.append(ClingoHeuristic(atom, type, bias, condition))

    # def add_minimize(self, priority: int, literals: List[Tuple[int, int]]) -> None:
    #     self.externals.append(ClingoMinimize(priority, literals))

    # def add_project(self, atoms: List[int]) -> None:
    #     self.externals.append(ClingoProject(atoms))

    # def add_weight_rule(self, head: List[int], lower: int, body: List[Tuple[int, int]], choice: bool = False) -> None:
    #     self.externals.append(ClingoWeightRule(choice, head, body, lower))


    # def as_list(self):
    #     return self.facts + self.rules + self.output_atoms

    # def __repr__(self):
    #     facts = '.\n'.join(repr(x) for x in self.facts)
    #     if facts:
    #         result = facts + '.'
    #     else:
    #         result = ''
    #     if self.cfacts:
    #         if result:
    #             result += '\n\n'
    #         result += '\n'.join(repr(x) for x in self.cfacts)
    #     if self.dfacts:
    #         if result:
    #             result += '\n\n'
    #         result += '\n'.join(repr(x) for x in self.dfacts)
    #     if self.rules:
    #         if result:
    #             result += '\n\n'
    #         result += '\n'.join(repr(x) for x in self.rules)
    #     # if self.output_atoms:
    #     #     if result:
    #     #         result += '\n\n'
    #     #     result += '\n'.join(repr(x) for x in self.output_atoms)
    #     if self.assumtions:
    #         if result:
    #             result += '\n\n'
    #         result += '\n'.join(repr(x) for x in self.assumtions)
    #     if self.externals:
    #         if result:
    #             result += '\n\n'
    #         result += '\n'.join(repr(x) for x in self.externals)
    #     if self.heuristics:
    #         if result:
    #             result += '\n\n'
    #         result += '\n'.join(repr(x) for x in self.heuristics)
    #     if self.minimizes:
    #         if result:
    #             result += '\n\n'
    #         result += '\n'.join(repr(x) for x in self.minimizes)
    #     if self.projections:
    #         if result:
    #             result += '\n\n'
    #         result += '\n'.join(repr(x) for x in self.projections)
    #     if self.weight_rules:
    #         result += '\n'.join(repr(x) for x in self.weight_rules)
    #     return result

class PrettyClingoOject:
    pass


class PrettyRule(PrettyClingoOject):

    def __init__(self, choice: bool = False, head: Iterable['Literal'] = [], body: Iterable['Literal'] = []): # pylint: disable=dangerous-default-value
        head = list(head)
        body = list(body)
        self.choice = choice
        self.head   = head
        self.body   = body

    def __repr__(self):
        if self.head:
            head = ', '.join(str(x) for x in self.head)
            if self.choice:
                head = '{ ' + head + ' }'
        else:
            head = ''
        if self.body:
            body = ':- ' + ', '.join(str(x) for x in self.body)
        else:
            body = ''
        if head and body:
            return head + ' ' + body + '.'
        else:
            return head + body + '.'

    def __lt__(self, other):
        if self.__class__ == other.__class__:
            return (self.choice, self.head, self.body) < (other.choice, other.head, other.body)
        # elif isinstance(other, ClingoObject):
        #     return self.order < other.order
        raise Exception("Incomparable type")


class PrettyProjection(PrettyClingoOject):

    def __init__(self, atoms: Iterable[Symbol]):
        if not isinstance(atoms, set):
            atoms = set(atoms)
        self.atoms = atoms

    def __repr__(self):
        atoms = ','.join(repr(atom) for atom in self.atoms)
        if atoms:
            return '#project ' + atoms  + '.'
        else:
            return '#project.'

    def __lt__(self, other):
        if self.__class__ == other.__class__:
            return self.atoms < other.atoms
        # elif isinstance(other, ClingoObject):
        #     return self.order < other.order
        raise Exception("Incomparable type")


class PrettyExternal(PrettyClingoOject):

    def __init__(self, atom: Symbol, value: TruthValue):
        self.atom  = atom
        self.value = value

    def __repr__(self):
        return '#external ' + repr(self.atom) + ' [' + ('True' if self.value else 'False') + '].'

class PrettyGroundProgram():

    def __init__(self, program: Iterable['ClingoObject']):
        self.symbols = set()
        self.atom_to_symbol: Dict[int, Symbol] = dict()
        self.facts:  List[Symbol]  = list()
        self.cfacts: List[Symbol]  = list()
        self.dfacts: List[Symbol]  = list()
        self.rules : List[PrettyRule]  = list()
        self.projections  : List[PrettyProjection] = list()
        self.assumtions   : List[ClingoAssume]     = list()
        self.externals    : List['ClingoExternal']         = list()
        self.heuristics   : List[ClingoHeuristic]  = list()
        self.minimizes    : List[ClingoMinimize]   = list()
        self.weight_rules : List[ClingoWeightRule] = list()
        self.add(program)

    def add(self, program: Iterable['ClingoObject']) -> None:
        output_atoms: Iterable[ClingoOutputAtom] = []
        others: Iterable['ClingoObject'] = []

        for obj in program:
            if isinstance(obj, ClingoOutputAtom):
                output_atoms.append(obj)
            else:
                others.append(obj)

        for output_atom in output_atoms:
            self.symbols.add(output_atom.symbol)
            if output_atom.atom != 0:
                self.atom_to_symbol.update({output_atom.atom : output_atom.symbol})
            else:
                self.facts.append(output_atom.symbol)

        self._add(others)
        # discard anonymous facts
        self.facts = list(fact for fact in self.facts if fact in self.symbols)


    def _rewrite_rule(self, rule: 'ClingoRule') -> PrettyRule:
        return self._rewrite_rule2(rule.choice, rule.head, rule.body)

    def _rewrite_rule2(self, choice: bool, head, body) -> PrettyRule:
        head = [self._rewrite_literal(literal) for literal in head]
        body = [self._rewrite_literal(literal) for literal in body]
        return PrettyRule(choice, head, body)

    def _rewrite_literal(self, literal) -> Literal:
        if abs(literal) in self.atom_to_symbol:
            lit = self.atom_to_symbol[abs(literal)]
        else:
            i = abs(literal)
            while True:
                lit = Function('x_' + str(i))
                if lit not in self.symbols:
                    break
                i += 1

        return Literal(lit, literal >= 0)

    def __add_rule(self, rule: PrettyRule) -> None:
        if not rule.body and len(rule.head) == 1:
            if rule.choice:
                self.cfacts.append(rule)
            else:
                self.facts.append(next(iter(rule.head)).atom)
        elif not rule.body:
            self.dfacts.append(rule)
        else:
            self.rules.append(rule)

    def add_rule(self, rule: 'ClingoRule') -> PrettyRule:
        rule = self._rewrite_rule(rule)
        self.__add_rule(rule)
        return rule

    def add_rules(self, rules: Iterable['ClingoRule']) -> None:
        for rule in rules:
            self.add_rule(rule)

    def _rewrite_projection(self, projection: 'ClingoProject') -> PrettyProjection:
        return PrettyProjection(self._rewrite_literal(atom) for atom in projection.atoms)

    def add_projection(self, projection: Union[PrettyProjection, 'ClingoProject']) -> None:
        projection = self._rewrite_projection(projection)
        self.projections.append(projection)

    def _rewrite_external(self, external: 'ClingoExternal') -> PrettyExternal:
        return PrettyExternal(self._rewrite_literal(external.atom), external.value)

    def add_external(self, external: 'ClingoExternal') -> None:
        external = self._rewrite_external(external)
        self.externals.append(external)

    def add_project(self, atoms: List[int]) -> None:
        self.add_projection(PrettyProjection(self.atom_to_symbol[atom] for atom in atoms))

    def _add(self, obj: Union['ClingoObject', Iterable['ClingoObject']]) -> None:
        if isinstance(obj, ClingoRule):
            self.add_rule(obj)
        if isinstance(obj, ClingoProject):
            self.add_projection(obj)
        elif isinstance(obj, ClingoAssume):
            self.assumtions.append(obj)
        elif isinstance(obj, ClingoExternal):
            self.add_external(obj)
        elif isinstance(obj, ClingoHeuristic):
            self.heuristics.append(obj)
        elif isinstance(obj, ClingoMinimize):
            self.minimizes.append(obj)
        elif isinstance(obj, ClingoWeightRule):
            self.weight_rules.append(obj)
        elif isinstance(obj, Iterable):
            for o in obj:
                self._add(o)

    def sort(self) -> None:
        self.facts.sort()
        self.cfacts.sort()
        self.dfacts.sort()
        self.rules.sort()
        self.projections.sort()
        self.assumtions.sort()
        self.externals.sort()
        self.heuristics.sort()
        self.minimizes.sort()
        self.weight_rules.sort()

    def __repr__(self):
        self.sort()
        facts = '.\n'.join(repr(x) for x in self.facts)
        if facts:
            result = facts + '.'
        else:
            result = ''
        if self.cfacts:
            if result:
                result += '\n\n'
            result += '\n'.join(repr(x) for x in self.cfacts)
        if self.dfacts:
            if result:
                result += '\n\n'
            result += '\n'.join(repr(x) for x in self.dfacts)
        if self.rules:
            if result:
                result += '\n\n'
            result += '\n'.join(repr(x) for x in self.rules)
        # if self.output_atoms:
        #     if result:
        #         result += '\n\n'
        #     result += '\n'.join(repr(x) for x in self.output_atoms)
        if self.assumtions:
            if result:
                result += '\n\n'
            result += '\n'.join(repr(x) for x in self.assumtions)
        if self.externals:
            if result:
                result += '\n\n'
            result += '\n'.join(repr(x) for x in self.externals)
        if self.heuristics:
            if result:
                result += '\n\n'
            result += '\n'.join(repr(x) for x in self.heuristics)
        if self.minimizes:
            if result:
                result += '\n\n'
            result += '\n'.join(repr(x) for x in self.minimizes)
        if self.projections:
            if result:
                result += '\n\n'
            result += '\n'.join(repr(x) for x in self.projections)
        if self.weight_rules:
            result += '\n'.join(repr(x) for x in self.weight_rules)
        return result


config = { 
    "decos" : "flow-cutter",
    "decot" : "1", 
    "knowledge_compiler" : "sharpsat-td", 
    "constrained" : "XD",
    "backdoort" : "30",
    "backdoors" : "fvs",
    "number_cores" : "1"
}


INTERNAL_ALGEBRAIC = "internal_algebraic"
EXTERNAL_ALGEBRAIC = "algebraic_atom"
INTERNAL_PENALTY = "internal_penalty"
EXTERNAL_PENALTY = "penalty_atom"

conversions = {
    INTERNAL_ALGEBRAIC : EXTERNAL_ALGEBRAIC,
    INTERNAL_PENALTY : EXTERNAL_PENALTY
}


import numpy as np
from aspmc.util import *

class ConstrainedDDNNF(object):
    # assumes that the DDNNF is smooth
    @staticmethod
    def parse_wmc(path, weights, cnf):
        """Performs nested algebraic model counting over an X/D-constrained circuit that is smooth while parsing it.
        
        Args:
            path (:obj:`string`): The path to the file that contains the circuit.
            weights (:obj:`string`): The weights of the literals. The weight for literal `v` is in `weights[2*(v-1)]`, the one for `-v` is in `weights[2*(v-1)+1]`
            cnf (:obj:`aspmc.compile.CNF`): The CNF which was compiled. Must contain the semirings, quantification partition (telling us which variables are to be interpreted over which semiring), and the transformation functions:
                transform (:obj:`string`): The transformation function used to transform a value from the second semiring into a value from the first semiring.
                    Will be used as

                            f_transform = eval(transform)
                            transform = lambda x : first_semiring.from_value(f_transform(x))
        Returns:
            (:obj:`object`): The algebraic model count.
        """
        shapes = [ (np.shape(weights[0])[0], ) + np.shape(semiring.one()) for semiring in cnf.semirings ]
        
        # transforms[i][j] for i < j transforms a value from the jth semiring into one from the jth semiring
        transforms = [ [ None for _ in range(len(cnf.semirings)) ] for _ in range(len(cnf.semirings)) ]
        # set the base transforms
        base_transforms = [ eval(transform) for transform in cnf.transforms ]
        base_transforms = [ lambda x, transform = transform, i = i: cnf.semirings[i].from_value(transform(x)) for i, transform in enumerate(base_transforms) ]

        for i in range(len(cnf.semirings) - 1):
            transforms[i][i+1] = base_transforms[i]


        # set the multistep transforms
        for i in range(len(cnf.semirings) - 1, -1, -1):
            for j in range(i - 2, -1, -1):
                transforms[j][i] = lambda x, i = i, j = j: base_transforms[j](transforms[j+1][i](x))
                
        var_to_val_type = [ len(cnf.semirings) - 1 for _ in range(cnf.nr_vars + 1) ]
        for val_type, vars in enumerate(cnf.quantified):
            for var in vars:
                var_to_val_type[var] = val_type

        with open(path) as ddnnf:
            _, nr_nodes, nr_edges, nr_leafs = ddnnf.readline().split()
            mem = []
            mem_types = []
            idx = 0
            for line in ddnnf:
                line = line.strip().split()
                if line[0] == 'L':
                    val = weights[to_pos(int(line[1]))]
                    val_type = var_to_val_type[abs(int(line[1]))]
                else:
                    if line[0] == 'A':
                        val = None
                        val_type = None
                        for child in line[2:]:
                            child = int(child)
                            if mem_types[child] != val_type:
                                if val_type is None:
                                    val_type = mem_types[child]
                                    val = np.empty(shapes[val_type], dtype=cnf.semirings[val_type].dtype)
                                    val[:] = cnf.semirings[val_type].one()
                                    val *= mem[child]
                                else:
                                    if mem_types[child] < val_type:
                                        val = np.array([ transforms[mem_types[child]][val_type](w) for w in val ], dtype = cnf.semirings[mem_types[child]].dtype)
                                        val_type = mem_types[child]
                                        val *= mem[child]
                                    else:
                                        val *= np.array([ transforms[val_type][mem_types[child]](w) for w in mem[child] ], dtype = cnf.semirings[val_type].dtype)
                            else:
                                val *= mem[child]
                    elif line[0] == 'O':
                        val = None
                        val_type = None
                        for child in line[3:]:
                            child = int(child)
                            if mem_types[child] != val_type:
                                if val_type is None:
                                    val_type = mem_types[child]
                                    val = np.empty(shapes[val_type], dtype=cnf.semirings[val_type].dtype)
                                    val[:] = cnf.semirings[val_type].zero()
                                    val += mem[child]
                                else:
                                    if mem_types[child] < val_type:
                                        val = np.array([ transforms[mem_types[child]][val_type](w) for w in val ], dtype = cnf.semirings[mem_types[child]].dtype)
                                        val_type = mem_types[child]
                                        val += mem[child]
                                    else:
                                        val += np.array([ transforms[val_type][mem_types[child]](w) for w in mem[child] ], dtype = cnf.semirings[val_type].dtype)
                            else:
                                val += mem[child]
                mem.append(val)
                mem_types.append(val_type)
                idx += 1
            return mem[idx - 1]




import signal
import psutil
import os

tempfiles = set()

def handler(signum, frame):
    for name in tempfiles:
        os.remove(name)
    # get pid of own process
    parent_pid = os.getpid()   
    parent = psutil.Process(parent_pid)
    # kill all children
    for child in parent.children(recursive=True):  
        child.kill()
    # kill self
    raise KeyboardInterrupt("Received Control+C, cleaning up and shutting down.")


signal.signal(signal.SIGINT, handler)
signal.signal(signal.SIGTERM, handler)


"""
Tree Decomposition module that introduces a tree decomposition class, a bag class and ways to construct tree decompositions.
"""

import networkx as nx
import subprocess
import inspect
import os
import queue
import logging
import time

src_path = os.path.abspath(os.path.realpath(inspect.getfile(inspect.currentframe())))
src_path = os.path.realpath(os.path.join(src_path, '../../external'))

logger = logging.getLogger("aspmc")

class Bag(object):
    """A class for bags of tree decompositions.

    Args:
        idx (:obj:`int`): The unique ID of this bag in the tree decomposition.
        vertices (:obj:`set`): The vertices in this bag.
        children (:obj:`list`): The children (also bags) of this bag in the tree decomposition.

    Attributes:
        idx (:obj:`int`): The unique ID of this bag in the tree decomposition.
        vertices (:obj:`set`): The vertices in this bag.
        children (:obj:`list`): The children (also bags) of this bag in the tree decomposition.
    """
    def __init__(self, idx, vertices, children):
        self.idx = idx
        self.vertices = vertices
        self.children = children

class TreeDecomposition(object):
    """A class for tree decompositions.

    Allows manipulation and iteration in post order.

    Args:
        bags (:obj:`int`): The number of bags this tree decomposition has.
        width (:obj:`int`): The width of this tree decomposition.
        vertices (:obj:`int`): The number of vertices in this tree decomposition.
        tree (:obj:`nx.Graph`): The graph representing the actual tree decomposition. 
            Must have a node attribute "bag" that contains all the vertices in the bag of the node.
        root (:obj:`int`, optional): Which node should be the root of the tree. Defaults to 1. 
        The root is only set if this argument is not `None`.


    Attributes:
        bags (:obj:`int`): The number of bags this tree decomposition has.
        width (:obj:`int`): The width of this tree decomposition.
        vertices (:obj:`int`): The number of vertices in this tree decomposition.
        tree (:obj:`nx.Graph`): The graph representing the actual tree decomposition. 
            Has a node attribute "bag" that contains the bag object for the node. 
            (Requires that `set_root` has been run already. This is typically done in the constructor if `root` is not `None`.)
        root (:obj:`int`): Which node is the root of the tree. 
    """
    def __init__(self, bags, width, vertices, tree, root = 1):
        self.bags = bags
        self.width = width
        self.vertices = vertices
        self.tree = tree
        self.root = None
        if root is not None:
            self.set_root(root)

    def get_root(self):
        """Return the bag of the root node.

        Returns:
            :obj:`Bag`: The bag of the root node.
        """
        return self.tree.nodes[self.root]["bag"]

    def get_bag(self, node):
        """Return the bag of given node.

        Args:
            node (:obj:`int`): The id of the node in the tree decomposition.

        Returns:
            :obj:`Bag`: The bag of the node.
        """
        return self.tree.nodes[node]["bag"]

    def set_root(self, root):
        """Sets the root of the tree decomposition.
        Updates the "bag" attributes of the tree so that the bags have the correct children.

        Args:
            root (:obj:`int`): The id of the root to be.

        Returns:
            None
        """
        first = self.root is None
        self.root = root
        stack = queue.LifoQueue()
        stack.put((-1, self.root, self.tree.neighbors(self.root)))
        while not stack.empty():
            parent, cur, neighbors = stack.get()
            try:
                neigh = next(neighbors)
                if neigh == parent:
                    neigh = next(neighbors)
                stack.put((parent, cur, neighbors))
                stack.put((cur, neigh, self.tree.neighbors(neigh)))
            except:
                if first:
                    vertices = self.tree.nodes[cur]["bag"]
                else:
                    vertices = self.tree.nodes[cur]["bag"].vertices
                self.tree.nodes[cur]["bag"] = Bag(cur, vertices, [ self.tree.nodes[x]["bag"] for x in self.tree.neighbors(cur) if x != parent ])

    def find_centroid(self):
        import sys
        if self.root is None:
            self.set_root(1)
        if self.bags == 1:
            return 1
        stack = queue.LifoQueue()
        stack.put((-1, self.root, self.tree.neighbors(self.root), 0))
        while not stack.empty():
            finished = True
            while finished:
                finished = False
                parent, cur, neighbors, count = stack.get()
                try:
                    neigh = next(neighbors)
                    if neigh == parent:
                        neigh = next(neighbors)
                    stack.put((parent, cur, neighbors, count))
                    stack.put((cur, neigh, self.tree.neighbors(neigh), 0))
                except StopIteration:
                    finished = True
                    here_count = len(self.get_bag(cur).vertices.difference(self.get_bag(parent).vertices))
                    here_count += count
                    pp, pc, pn, pcount = stack.get()
                    if here_count + pcount > self.vertices//2:
                        return pc
                    stack.put((pp, pc, pn, pcount + here_count))


    def __iter__(self):
        if self.root is None:
            self.set_root(1)
        stack = queue.LifoQueue()
        stack.put((-1, self.root, self.tree.neighbors(self.root)))
        while not stack.empty():
            parent, cur, neighbors = stack.get()
            try:
                neigh = next(neighbors)
                if neigh == parent:
                    neigh = next(neighbors)
                stack.put((parent, cur, neighbors))
                stack.put((cur, neigh, self.tree.neighbors(neigh)))
            except:
                yield cur

    def __str__(self):
        res = f"s tw {self.bags} {self.width} {self.vertices}\n"
        for bag in self.bag_iter():
            res += f"b {bag.idx} {' '.join([ str(v) for v in bag.vertices ])}\n"
        for bag in self.bag_iter():
            for neigh in bag.children:
                res += f"{bag.idx} {neigh.idx}\n"
        return res

    def bag_iter(self, order = "post-order"): 
        """Iterate over all the bags in post order.

        Args:
            order (:obj:`string`, optional): Which order the bags should be iterated over. 
            Currently supported are "post-order" and "pre-order". Defaults to "post-order".

        Returns:
            :obj:`iterator`: Yields over the bags in the specified order.
        """    
        if order == "post-order":
            for bag in self:
                yield self.tree.nodes[bag]["bag"]
        elif order == "pre-order":
            if self.root is None:
                self.set_root(1)
            stack = queue.LifoQueue()
            stack.put((-1, self.root, self.tree.neighbors(self.root)))
            yield self.get_bag(self.root)
            while not stack.empty():
                parent, cur, neighbors = stack.get()
                try:
                    neigh = next(neighbors)
                    if neigh == parent:
                        neigh = next(neighbors)
                    stack.put((parent, cur, neighbors))
                    stack.put((cur, neigh, self.tree.neighbors(neigh)))
                    yield self.get_bag(neigh)
                except:
                    continue
        else:
            logger.error(f"Unsupported order {order}.")
            exit(-1)

    def find_containing(self, vertices):
        """Find a bag containing all the vertices in `vertices`.

        Args:
            vertices (:obj:`set`): The vertices that should be contained in the desired bag.

        Returns:
            :obj:`Bag`: A bag containg all the vertices in `vertices` or `None` if there is no such bag.
        """
        for t in self:
            if self.tree.nodes[t]["bag"].vertices.issuperset(vertices):
                return t

    def remove(self, vertices):
        """Remove the vertices in `vertices` from all bags.

        Args:
            vertices (:obj:`set`): The vertices that should be removed.

        Returns:
            None
        """
        for t in self:
            self.tree.nodes[t]["bag"].vertices.difference_update(vertices)
                
    def draw(self):
        """Visualizes this tree decompisition by showing a plot of it. 
        Mostly useful for debugging.

        Returns:
            None
        """
        import matplotlib.pyplot as plt
        from networkx.drawing.nx_pydot import graphviz_layout
        labels = { v : str((v, self.tree.nodes[v]["bag"])) for v in self.tree.nodes }
        pos = graphviz_layout(self.tree, prog="dot")
        nx.draw(self.tree, pos)
        nx.draw_networkx_labels(self.tree, pos, labels)
        plt.tight_layout()
        plt.axis("off")
        plt.show()


                


def from_file(path):
    """Reads a tree decomposition from a file.

    Args:
        path (:obj:`string`): The path to the file containing the tree decomposition.

    Returns:
        :obj:`TreeDecomposition`: The tree decomposition specified in the file.
    """
    with open(path, "r") as input:
        for line in input:
            line = line.split()
            if len(line) == 0: 
                continue
            if line[0] == 's':
                bags = int(line[2])
                width = int(line[3])
                vertices = int(line[4])
                tree = nx.Graph()
                tree.add_nodes_from(range(1,bags + 1))
            elif line[0] == 'b':
                tree.nodes[int(line[1])]["bag"] = set([ int(x) for x in line[2:] ])
            else:
                tree.add_edge(int(line[0]), int(line[1]))
    return TreeDecomposition(bags, width, vertices, tree)

def from_stream(stream):
    """Reads a tree decomposition from a stream.

    Args:
        stream (:obj:`string`): The stream with the tree decomposition. Must be binary.

    Returns:
        :obj:`TreeDecomposition`: The tree decomposition specified in the stream.
    """
    for line in stream.read().decode().split('\n'):
        line = line.split()
        if len(line) == 0: 
            continue
        if line[0] == 's':
            bags = int(line[2])
            width = int(line[3])
            vertices = int(line[4])
            tree = nx.Graph()
            tree.add_nodes_from(range(1,bags + 1))
        elif line[0] == 'b':
            tree.nodes[int(line[1])]["bag"] = set([ int(x) for x in line[2:] ])
        else:
            tree.add_edge(int(line[0]), int(line[1]))
    return TreeDecomposition(bags, width, vertices, tree)

def from_graph(graph, solver = "flow-cutter", timeout = "1"):
    """Constructs a tree decomposition from a graph.

    Args:
        solver (:obj:`string`, optional): The solver to generate the tree decomposition. 
        Currently supported is "flow-cutter". Defaults to "flow-cutter".
        timeout (:obj:`string`, optional): After how much time to kill the solver in seconds. Defaults to "1".

    Returns:
        :obj:`TreeDecomposition`: The tree decomposition output by the solver.
    """
    timeout = float(timeout)
    start = time.time()
    node_map = dict()
    map_node = dict()
    idx = 1
    for node in graph.nodes:
        node_map[node] = idx
        map_node[idx] = node
        idx += 1
    if solver == "flow-cutter":
        p = subprocess.Popen([os.path.join(src_path, "flow-cutter/flow_cutter_pace17")], stdin=subprocess.PIPE, stdout=subprocess.PIPE, close_fds = True)
    else:
        logger.error(f"Unknown td-solver {solver}")    
    # with open("debug.gr", "wb") as debug_gr:
    #    debug_gr.write(f"p tw {len(graph.nodes)} {len(graph.edges)}\n".encode())
    #    for (v,vp) in graph.edges:
    #        debug_gr.write(f"{node_map[v]} {node_map[vp]}\n".encode())
    p.stdin.write(f"p tw {len(graph.nodes)} {len(graph.edges)}\n".encode())
    for (v,vp) in graph.edges:
        p.stdin.write(f"{node_map[v]} {node_map[vp]}\n".encode())
    p.stdin.flush()
    p.stdin.close()
    first = False
    while p.poll() is None and not first:
        line = p.stdout.readline().decode()
        first = line.startswith("c status")
    time_left = timeout - (time.time() - start)
    try:
        p.wait(time_left)
    except subprocess.TimeoutExpired:
        p.terminate()
    for line in p.stdout.read().decode().split('\n'):
        line = line.split()
        if len(line) == 0: 
            continue
        if line[0] == 's':
            bags = int(line[2])
            width = int(line[3])
            vertices = int(line[4])
            tree = nx.Graph()
            tree.add_nodes_from(range(1,bags + 1))
        elif line[0] == 'b':
            tree.nodes[int(line[1])]["bag"] = set([ map_node[int(x)] for x in line[2:] ])
        elif line[0] == 'c':
            continue
        else:
            tree.add_edge(int(line[0]), int(line[1]))
    p.stdout.close()
    p.wait()
    return TreeDecomposition(bags, width, vertices, tree)

def from_hypergraph(hypergraph, solver = "flow-cutter", timeout = "1"):
    """Constructs a tree decomposition from a hypergraph.

    Args:
        solver (:obj:`string`, optional): The solver to generate the tree decomposition. 
        Currently supported is "flow-cutter". Defaults to "flow-cutter".
        timeout (:obj:`string`, optional): After how much time to kill the solver in seconds. Defaults to "1".

    Returns:
        :obj:`TreeDecomposition`: The tree decomposition output by the solver.
    """
    if solver == "flow-cutter":
        return from_graph(hypergraph.to_graph(), solver = solver, timeout = timeout)
    else:
        logger.error(f"Unknown td-solver {solver}")   








class MaxPlusGradient(object):
    def __init__(self, value):
        self.value = value

    def __add__(self, other):
        return self if self.value[1] >= other.value[1] else other

    def __iadd__(self, other):
        self.value = self.value if self.value[1] >= other.value[1] else other
        return self

    def __mul__(self, other):
        return MaxPlusGradient((self.value[0] * other.value[0], self.value[0] * other.value[1] + self.value[1] * other.value[0]))

    def __imul__(self, other):
        self.value[1] *= other.value[0]
        self.value[1] += self.value[0] * other.value[1]
        self.value[0] *= other.value[0]
        return self

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(self)

def parse(value, atom = None):
    value = value[1:-1]
    value = value.split(',')
    return MaxPlusGradient((float(value[0]), float(value[1])))

def from_value(value):
    return MaxPlusGradient(value)

def negate(value):
    return MaxPlusGradient((1.0 - value.value[0], -value.value[1]))

def to_string(value):
    return f"({value.value[0]},{value.value[1]})"
    
def is_idempotent():
    return True

def zero():
    return MaxPlusGradient((0.0, float("-inf")))

def one():
    return MaxPlusGradient((1.0, 0.0))

dtype = object
pattern = '[(][+-]?([0-9]*[.])?[0-9]+[,][+-]?([0-9]*[.])?[0-9]+[)]'


from lark import Lark, Transformer
import re

from sympy import true

from aspmc.parsing.lark_parser import Atom

class NormalRule(object):
    """A class for normal rules.

    If a rule has more than one atom in the head, the head must be an annotated disjunction.
    Then each of the atoms must have a weight.

    Implements a custom `__str__` method.

    Args:        
        head (:obj:`list`): The list of head atoms. May be empty.
        body (:obj:`list`): The list of body atoms. May be empty.
        choice (bool): Whether the rule is a choice rule.

    Attributes:
        head (:obj:`list`): The list of head atoms. May be empty.
        body (:obj:`list`): The list of body atoms. May be empty.
        choice (bool): Whether the rule is a choice rule.
    """
    def __init__(self, head, body, choice):
        self.head = head
        self.body = body if body is not None else []
        self.choice = choice

    def __str__(self):
        res = ""
        if self.head is not None:
            if self.choice:
                res += f"{{{str(self.head[0])}}}"
            else:
                res += f"{str(self.head[0])}"
        if len(self.body) > 0:
            res +=f":-{','.join([str(x) for x in self.body])}."
        else:
            res += "."
        return res

    def __repr__(self):
        return str(self)
    
    def asp_string(self):
        """Generates an ASP representation of the rule.

        Implements a custom `__str__` method.
        
        Returns:
            :obj:`string`: The representation of this rule as an ASP rule.
        """
        return str(self)
    
class WeakConstraint(object):
    """A class for weak constraints.
    
    Weak constraints are constraints whose satisfaction is (un)desirable. 
    This means they have a weight and an empty head.
    Furthermore, every weak constraint may have a list of terms associated with it. 
    The penalty `weight` will be triggered exactly once if any weak constraint with this list of terms
    is not satisfied.

    Implements a custom `__str__` method.

    Args:        
        body (:obj:`list`): The list of body atoms. May be empty.
        weight (int): Whether the rule is a choice rule.
        terms (:obj:`list`): The list of terms that associated with this constraint. May be empty.

    Attributes:
        body (:obj:`list`): The list of body atoms. May be empty.
        weight (int): Whether the rule is a choice rule.
        terms (:obj:`list`): The list of terms that associated with this constraint. May be empty.
    """
    def __init__(self, body, weight, terms):
        self.body = body
        self.weight = weight
        self.terms = terms if terms is not None else []

    def __str__(self):
        res = ":~"
        if len(self.body) > 0:
            res +=f"{','.join([str(x) for x in self.body])}."
        else:
            res += "."
        res += f"[{self.weight}{'' if len(self.terms) == 0 else ','}{','.join([str(x) for x in self.terms])}]"
        return res

    def __repr__(self):
        return str(self)
    
    def asp_string(self):
        """Generates an ASP representation of the rule.

        Implements a custom `__str__` method.
        
        Returns:
            :obj:`string`: The representation of this rule as an ASP rule.
        """
        return str(self)


class Atom(object):
    """A class for atoms.

    Implements a custom `__str__` method.
    
    Args:
        predicate (:obj:`string`): The predicate of the atom.
        inputs (:obj:`list`, optional): The inputs of the atom. 
        These may be strings or other atoms. 
        Defaults to `None`.
        negated (:obj:`bool`, optional): Whether the atom is negated.
        Defaults to `False`.

    Attributes:
        predicate (:obj:`string`): The predicate of the atom.
        inputs (:obj:`list`, optional): The inputs of the atom. 
        These may be strings or other atoms. 
        negated (:obj:`bool`, optional): Whether the atom is negated.
    """
    def __init__(self, predicate, inputs = None, negated=False):
        self.predicate = predicate
        self.inputs = inputs if inputs is not None else []
        def replace_quotes(term):
            if type(term) != Atom:
                return term.replace("'", '"')
            return term
        self.inputs = [ replace_quotes(term) for term in self.inputs ]
        self.negated = negated

    def __str__(self):
        res = ""
        if self.negated:
            res += "not "
        res += f"{self.predicate}"
        if len(self.inputs) > 0:
            res += f"({','.join([ str(term) for term in self.inputs ])})"
        return res

    def __repr__(self):
        return str(self)

    def get_variables(self):
        """Rcursively finds all the variables used in the atom.

        Returns:
            :obj:`list`: The list of variables as strings.
        """
        vars = set()
        for term in self.inputs:
            if type(term) == Atom:
                vars.update(term.get_variables())
            elif re.match(r"[A-Z][a-zA-Z0-9]*", term):
                vars.add(term)
        return vars


class OptProgramTransformer(Transformer):
    """The corresponding OptProgram semantics class for the OPTGRAMMAR grammar.
    
    See the lark documentation for how this works.
    """
    def program(self, ast):  # noqa
        return ast # sort out the comments

    def rule(self, ast):  # noqa
        return ast[0]

    def fact(self, ast): #noqa
        ast = ast[0]
        return NormalRule([ast['atom']], [], ast['choice'])

    def normal_rule(self, ast):  # noqa
        return NormalRule([ast[0]['atom']], ast[1], ast[0]['choice'])
    
    def constraint(self, ast): #noqa
        return NormalRule(None, ast[0], False)
    
    def weakconstraint(self, ast): #noqa
        return WeakConstraint(ast[0], ast[1], ast[2])
    
    def head(self, ast):
        if len(ast) == 3:
            return { 'atom' : ast[1], 'choice' : True }
        else:
            return { 'atom' : ast[0], 'choice' : False }
    
    def body(self, ast):  # noqa
        if len(ast) == 1 and ast[0] == None:
            return None
        return ast

    def atom(self, ast):  # noqa
        negated = str(ast[0]) == 'not'
        if len(ast) == 3:
            return Atom(str(ast[1]), inputs = ast[2], negated = negated)
        else:
            return Atom(str(ast[1]), negated = negated)

    def input(self, ast):  # noqa
        return ast

    def term(self, ast):  # noqa
        ast = ast[0]
        if type(ast) == Atom:
            return ast
    
        if "." in ast and (ast[0] != '"' or ast[-1] != '"'):
            return '"' + ast + '"'
        return str(ast)

    def variable(self, ast): # noqa
        return str(ast[0])

    def weight(self, ast):  # noqa
        return str(ast[0])


GRAMMAR = r'''
    program : rule*

    rule : normal_rule | fact | constraint | weakconstraint

    fact : head "."

    normal_rule : head ":-" body  "."

    constraint : ":-" body  "."
    
    weakconstraint : ":~" body  "." "[" weight ["," input] "]"

    head : ( atom | (OPEN_ANGLE atom CLOSED_ANGLE) )

    body : [ atom ( "," atom )* ]

    OPEN_ANGLE : "{"
    
    CLOSED_ANGLE : "}"
    
    NEGATION : "not"

    atom : [NEGATION] ( /[a-z]([a-zA-Z0-9_])*/ [ "(" input ")" ]  |  "(" /[a-z]([a-zA-Z0-9_])*/ [ "(" input ")" ] ")" )

    input : term ( "," term )*

    term : atom | /[0-9_\/<>=+"-]([a-zA-Z0-9_\/<>=+".-]*)/ | variable 

    variable : /[A-Z][a-zA-Z0-9]*/

    weight :  /[+-]?[1-9]+[0-9]*/ | variable

    COMMENT : "%" /[^\n]+/
    %ignore COMMENT
    %import common.WS
    %ignore WS
    
'''


if __name__ == '__main__':
    import sys
    parser = Lark(GRAMMAR, start='program', parser='lalr', transformer=OptProgramTransformer())

    with open(sys.argv[1]) as infile:
        tree = parser.parse(infile.read())
        for r in tree:
            print(r)


"""
Program module providing the algebraic progam class.
"""

import re
import numpy as np
import logging

from aspmc.programs.program import Program, UnsupportedException

from aspmc.parsing.clingoparser.clingoext import ClingoRule, Control

from lark import Lark
from aspmc.parsing.lark_parser import GRAMMAR, ProblogTransformer
import aspmc.programs.grounder as grounder

from aspmc.util import *

logger = logging.getLogger("aspmc")

class AlgebraicProgram(Program):
    """A class for programs with weights over semirings. 

    Should be specified in ProbLog syntax, but allows negations and negative cycles.

    Annotated disjunctions are theoretically supported over any semiring but the results are likely to 
    be wrong over semirings that differ from the probabilistic semiring.

    Grounding of these programs (and subclasses thereof) should follow the following strategy:

    * `_prepare_grounding(self, program)` should take the output of the parser 
        (i.e. a list of rules and special objects) and process all the rules and special objects
        transforming them either into other rules or into strings that can be given to the grounder.
    * the output of `_prepare_grounding(self, program)` is transformed to one program string via

            '\\n'.join([ str(r) for r in program ])
        
        This string will be given to the grounder, which produces a clingo control object.
    * `_process_grounding(self, clingo_control)` should take this clingo control object and process the
        grounding in an appropriate way (and draw some information from it optionally about weights, special objects).
        The resulting processed clingo_control object must only know about the 
        rules that should be seen in the base program class.

    Thus, subclasses can override `_prepare_grounding` and `_process_grounding` (and optionally call the superclass methods) 
    to handle their extras. See aspmc.programs.meuprogram or aspmc.programs.smprogram for examples.

    Args:
        program_str (:obj:`string`): A string containing a part of the program in ProbLog syntax. 
        May be the empty string.
        program_files (:obj:`list`): A list of string that are paths to files which contain programs in 
        ProbLog syntax that should be included. May be an empty list.
        semiring (:obj:`module`): The semiring module to be used. See aspmc.semiring for how they should look.

    Attributes:
        semiring (:obj:`module`): The semiring module to be used. 
        weights (:obj:`dict`): The dictionary from atom names to their weight.
        queries (:obj:`list`): The list of atoms to be queries in their string representation.
    """
    def __init__(self, program_str, program_files, semiring):
        self.semiring = semiring
        self.weights = {}
        self.queries = []
        self.annotated_disjunctions = []
        for path in program_files:
            with open(path) as file_:
                program_str += file_.read()
        # parse
        my_grammar = GRAMMAR + f"%override weight : /{self.semiring.pattern}/ | variable\n"
        parser = Lark(my_grammar, start='program', parser='lalr', transformer=ProblogTransformer())
        program = parser.parse(program_str)

        # ground
        clingo_control = Control()
        self._ground(clingo_control, program)

        # initialize the superclass
        Program.__init__(self, clingo_control = clingo_control)

    def _ground(self, clingo_control, program):
        # do the grounding in three steps:
        # 1. transform the parsed rules into asp rules
        program = self._prepare_grounding(program)
        # 2. give the asp rules to the grounder
        clingo_str = '\n'.join([ str(r) for r in program ])
        grounder.ground(clingo_control, program_str = clingo_str, program_files = [])
        # 3. take care of possible extras
        self._process_grounding(clingo_control)

    def _prepare_grounding(self, program):
        guess_idx = 0
        new_program = []
        for r in program:
            if isinstance(r, str):
                new_program.append(r)
            elif r.head is not None and r.head[0].predicate == "query":
                atom = r.head[0].inputs[0]
                new_program.append(f"query_atom({atom}):-{atom}.")
            elif r.weights is not None:
                self.annotated_disjunctions.append({})
                variables = set()
                for atom in r.body:
                    variables.update(atom.get_variables())
                if len(variables) == 0:
                    variables.add("none")
                variables = f"set({','.join([ str(x) for x in variables ])})"
                # generate the head, which guesses at most one of the algebriac atoms
                # at the same time generate the rules, which use these atoms
                guess_atoms = []
                for i in range(len(r.weights)):
                    # the guess atom
                    guess_atom = f"algebraic_atom({(guess_idx, i)},{variables},{r.head[i]},"
                    if re.match(r"[A-Z][a-zA-Z0-9]*", str(r.weights[i])):
                        guess_atom += str(r.weights[i]) + ")"
                    else:
                        guess_atom += f"\"{r.weights[i]}\")"
                    guess_atoms.append(guess_atom)
                    # add the rule which uses this guess
                    use_rule = f"{r.head[i]}:-{','.join([str(x) for x in r.body]+[guess_atom])}."
                    new_program.append(use_rule)
                # finally add an atom for the case that none of the other atoms are true
                guess_atom = f"algebraic_atom({(guess_idx, i+1)},{variables},none,\"none\")"
                guess_atoms.append(guess_atom)
                # the rule that ensures that exactly one or none of the atoms are true
                guess_rule = f"{';'.join(guess_atoms)}:-{','.join([str(x) for x in r.body])}."
                new_program.append(guess_rule)
                self.annotated_disjunctions[guess_idx]
                guess_idx += 1
            else:
                new_program.append(r)
        return new_program

    def _process_grounding(self, clingo_control):
        new_objects = []
        symbol_map = {}
        conditioned = {}
        for sym in clingo_control.symbolic_atoms:
            symbol_map[sym.literal] = str(sym.symbol)
        for o in clingo_control.ground_program.objects:
            if isinstance(o, ClingoRule):
                if len(o.head) == 0:
                    head_name = ""
                else:
                    if abs(o.head[0]) in symbol_map:
                        head_name = symbol_map[abs(o.head[0])]
                    else:
                        head_name = ""
                # check if the rule corresponds to a query
                if head_name.startswith("query_atom"):
                    if len(o.body) > 0:
                        self.queries.append(symbol_map[abs(o.body[0])])
                    else:
                        atom = head_name[11:-1]
                        logger.warning(f"Query for atom {atom} was proven true during grounding.")
                        logger.warning(f"Including it has a negative impact on performance.")
                        self.queries.append(atom)
                elif head_name.startswith("algebraic_atom"):
                    # find out the index of the annotated disjunction
                    cur_idx = 16
                    next_idx = head_name.find(",", cur_idx)
                    guess_idx = int(head_name[cur_idx:next_idx])
                    cur_idx = next_idx + 1
                    next_idx = head_name.find(")", cur_idx)
                    # find out which terms were used for grounding
                    cur_idx = next_idx + 6 # == len("),set(")
                    end_idx = head_name.find(')', cur_idx)
                    # now cur_idx is the location of the closing bracket
                    variables = tuple(head_name[cur_idx: end_idx].split(","))
                    self.annotated_disjunctions[guess_idx][variables] = o
                # check if the rule has algebraic_atom in the body
                # these rules are handled later
                elif any([ a > 0 and a in symbol_map and symbol_map[a].startswith("algebraic_atom") for a in o.body ]):
                    for a in o.body:
                        if a > 0 and a in symbol_map and symbol_map[a].startswith("algebraic_atom"):
                            conditioned[a] = o
                else:
                    new_objects.append(o)

        for idx in range(len(self.annotated_disjunctions)):
            for variables in self.annotated_disjunctions[idx]:
                guess_rule = self.annotated_disjunctions[idx][variables]
                sum_all = self.semiring.zero()
                guess_rule.body = [] # make the guess unconditional
                new_objects.append(guess_rule)
                for atom in guess_rule.head:
                    # add all the rules that are conditioned on an algebraic atom
                    if atom in conditioned:
                        new_objects.append(conditioned[atom])
                    # find out the weights
                    head_name = symbol_map[atom]
                    start = len(head_name) - 3
                    while head_name[start] != "\"":
                        start -= 1
                    start += 1
                    if head_name[start:-2] == "none":
                        none_atom = atom
                        continue
                    weight = self.semiring.parse(head_name[start:-2])
                    self.weights[(head_name, True)] = weight
                    # no need to set the weight of the negation to one, since this is the standard
                    sum_all += weight
                # assign the atom that is true when none of the other atoms are true the negation of the sum of the weights
                head_name = symbol_map[none_atom]
                self.weights[(head_name, True)] = self.semiring.negate(sum_all)

        clingo_control.ground_program.objects = new_objects

    def _prog_string(self, program):
        result = ""
        for r in self._exactlyOneOf:
            result += ";".join([ f"{self.weights[(self._internal_name(v),True)]}::{self._external_name(v)}" for v in r ])
            result += ".\n"
        for r in program:
            result += ";".join([self._external_name(v) for v in r.head])
            result += ":-"
            result += ",".join([("not " if v < 0 else "") + self._external_name(abs(v)) for v in r.body])
            result += ".\n"
        for query in self.queries:
            result += f"query({query}).\n"
        return result

    def _finalize_cnf(self):
        weight_list = self.get_weights()
        for v in range(self._cnf.nr_vars*2):
            self._cnf.weights[to_dimacs(v)] = weight_list[v]
        self._cnf.semirings = [ self.semiring ]
        self._cnf.quantified = [ list(range(1, self._cnf.nr_vars + 1)) ]
        query_cnt = max(len(self.queries), 1)
        for v in self._deriv:
            if all(self._cnf.weights[v][i] == self.semiring.one() and self._cnf.weights[-v][i] == self.semiring.one() for i in range(query_cnt)):
                self._cnf.auxilliary.add(v)

    def get_weights(self):
        query_cnt = max(len(self.queries), 1)
        varMap = { name : var for var, name in self._nameMap.items() }
        weight_list = [ np.full(query_cnt, self.semiring.one(), dtype=self.semiring.dtype) for _ in range(self._cnf.nr_vars*2) ]
        for (name, phase) in self.weights:
            if phase:
                weight_list[to_pos(varMap[name])] = np.full(query_cnt, self.weights[(name, phase)], dtype=self.semiring.dtype)
            else:
                weight_list[neg(to_pos(varMap[name]))] = np.full(query_cnt, self.weights[(name, phase)], dtype=self.semiring.dtype)
        for i, query in enumerate(self.queries):
            weight_list[neg(to_pos(varMap[query]))][i] = self.semiring.zero()
        return weight_list

    def get_queries(self):
        return self.queries



#!/usr/bin/env python3

"""
Maximum A Posteriori Program module specializing the TwoAlgebraicProgam class.
"""
import logging

from aspmc.parsing.clingoparser.clingoext import Control

from aspmc.programs.problogprogram import ProblogProgram
from aspmc.programs.twoalgebraicprogram import TwoAlgebraicProgram
from lark import Lark
from aspmc.parsing.lark_parser import GRAMMAR, ProblogTransformer
from aspmc.util import *

import aspmc.semirings.maxtimesdecisions as first_semiring
import aspmc.semirings.probabilistic as second_semiring

logger = logging.getLogger("aspmc")

class MAPProblogProgram(TwoAlgebraicProgram, ProblogProgram):
    """A class for Maximum A Posteriori programs. 

    The syntax for these programs is the same as for ProbLog in MAP mode. 
    This means that the probabilistic part is specified as usual,
    evidence atoms `a`, i.e. atoms that are true, must be given as `evidence(a).` and
    the MAP query atoms `q`, whose assignment should be maximized over should be given as `query(q).`.

    Note that these atoms can be any atom and need not be probabilistic facts.

    Subclasses `TwoAlgebraicProgram` since it is a second level problem.

    Overrides the `_prepare_grounding` method to deal with MAP query atoms and evidence.

    Args:
        program_str (:obj:`string`): A string containing a part of the program in MAP ProbLog syntax. 
        May be the empty string.
        program_files (:obj:`list`): A list of string that are paths to files which contain programs in 
        MAP ProbLog syntax that should be included. May be an empty list.
    """
    def __init__(self, program_str, program_files):
        self.semiring = second_semiring
        self.weights = {}
        self.evidence = {}
        self.map_variables = []
        self.queries = []
        self.annotated_disjunctions = []
        for path in program_files:
            with open(path) as file_:
                program_str += file_.read()
        # parse the program
        my_grammar = GRAMMAR + f"%override weight : /{self.semiring.pattern}/ | variable\n"
        parser = Lark(my_grammar, start='program', parser='lalr', transformer=ProblogTransformer())
        program = parser.parse(program_str)

        # ground the program
        clingo_control = Control()
        self._ground(clingo_control, program)

        self.map_variables = self.queries
        self.queries = []

        # set the weights
        first_weight_list = {}
        first_semiring.names = [ name for name in self.map_variables ]
        # weights for the maxtimes semiring
        for i, name in enumerate(self.map_variables):
            if name in self.weights:
                first_weight_list[(name, True)] = first_semiring.MaxTimesFloat(self.weights[name], 2**i)
                first_weight_list[(name, False)] = first_semiring.MaxTimesFloat(1.0 - self.weights[name], 0)
            else:
                first_weight_list[(name, True)] = first_semiring.MaxTimesFloat(1.0, 2**i)
                first_weight_list[(name, False)] = first_semiring.MaxTimesFloat(1.0, 0)
        # weights for the probability semiring
        second_weight_list = {}
        for name in self.weights:
            if name in self.map_variables:
                continue
            # # we only want the weights for those probabilistic atoms which are *not* map_variables
            # # get the original atom name
            # cur_idx = 16
            # next_idx = name.find(")", cur_idx)
            # cur_idx = next_idx + 6 # == len("),set(")
            # next_idx = name.find(')', cur_idx)
            # cur_idx = next_idx + 2 # == len("),")
            # start_idx = cur_idx
            # open_brackets = 0
            # open_quotes = False
            # while name[cur_idx] != ',' or open_brackets > 0 or open_quotes:
            #     if name[cur_idx] == '"':
            #         open_quotes = not open_quotes
            #     elif name[cur_idx] == '(':
            #         open_brackets += 1
            #     elif name[cur_idx] == ')':
            #         open_brackets -= 1
            #     cur_idx += 1
            # end_idx = cur_idx
            # if True or not name[start_idx:end_idx] in self.map_variables:
            second_weight_list[(name, True)] = self.weights[name]
            second_weight_list[(name, False)] = 1.0 - self.weights[name]
        TwoAlgebraicProgram.__init__(self, clingo_control, first_semiring, second_semiring, first_weight_list, second_weight_list, "lambda w : w", self.queries)

    def _prepare_grounding(self, program):
        new_program = []
        for r in program:
            if isinstance(r, str):
                new_program.append(r)
            elif r.head is not None and r.head[0].predicate == "evidence":
                atom = r.head[0].inputs[0]
                new_program.append(f":-{'' if atom.negated else 'not'} {str(atom)[4:] if atom.negated else str(atom)}.")
                phase = atom.negated
                atom.negated = False
                self.evidence[str(atom)] = phase
            else:
                new_program.append(r)
        return super()._prepare_grounding(new_program)

    def _prog_string(self, program):
        result = ""
        for v in self._guess:
            result += f"{self.weights[self._internal_name(v)]}::{self._external_name(v)}.\n"
        for name in self.evidence:
            result += "evidence("
            if self.evidence[name]:
                result += "\\+"
            result += f"{name}).\n"
        for r in program:
            if not r.head:
                continue
            result += ";".join([self._external_name(v) for v in r.head])
            if len(r.body) > 0:
                result += ":-"
                result += ",".join([("\\+ " if v < 0 else "") + self._external_name(abs(v)) for v in r.body])
            result += ".\n"
        for query in self.map_variables:
            result += f"query({query}).\n"
        return result

    def to_pita(self):
        result = """:- use_module(library(pita)).
:- pita.
:- begin_lpad.
"""
        for v in self._guess:
            if self._external_name(v) in self.map_variables:
                result += "map_query "
            result += f"{self.weights[self._internal_name(v)]}::{self._external_name(v)}.\n"
        if self.evidence:
            result += "ev :- "
            result += ",".join([("\\+" if self.evidence[name] else "") + name for name in self.evidence ])
            result += ".\n"
        for r in self._program:
            if not r.head:
                continue
            result += ";".join([self._external_name(v) for v in r.head])
            if len(r.body) > 0:
                result += ":-"
                result += ",".join([("\\+ " if v < 0 else "") + self._external_name(abs(v)) for v in r.body])
            result += ".\n"
        result += ":- end_lpad."
        return result


import sys
import clingo
import threading
import logging

from aspmc.config import config


logger = logging.getLogger("aspmc/backdoor_search")
logging.basicConfig(format='[%(levelname)s] %(name)s: %(message)s', level="INFO")

def safe_int(string):
    try:
        return int(string)
    except ValueError:
        return string

class ClingoControl:
    def __init__(self, prog):
        self.prog = prog
        self.grounded = False
        self.timeout = False

    # this method tries integer graph nodes if possible
    # encodingFile - encoding to use, probably guess_v2.lp is the best so far, prevents direct encoding of reachability
    # timeout - seconds to wait at most for result
    # solve_limit, see clingo documentation
    def get_backdoor(self, encodingFile, timeout=30, usc=False, solve_limit="umax,umax"):
        if clingo is None:
            raise ImportError()

        aset = [sys.maxsize, False, [], None, []]
        
        def __on_model(model):
            #if len(model.cost) == 0:
            #    return
            
            logger.debug("better answer set found: %s %s %s", model, model.cost, model.optimality_proven)
            
            aset[1] |= model.optimality_proven
            opt = model.cost[0] if len(model.cost) > 0 else 0
            if opt <= aset[0]:
                if opt < aset[0]:
                    aset[2] = []
                aset[0] = opt
                answer_set = [safe_int(x) for x in str(model).translate(str.maketrans(dict.fromkeys("abs()"))).split(" ")]
                if len(answer_set) == 1 and answer_set[0] == '':
                    answer_set = []

                # might get "fake" duplicates :(, with different model.optimality_proven
                if answer_set not in aset[2][-1:]:
                    aset[2].append(answer_set)
        if encodingFile != None:
            with open(encodingFile,"r") as encoding:
                encodingContent = "".join(encoding.readlines())

            # FIXME: use mutable string
            self.prog += encodingContent
        
        c = clingo.Control()
        if usc:
            c.configuration.solver.opt_strategy = "usc,pmres,disjoint,stratify"
            c.configuration.solver.opt_usc_shrink = "min"
        c.configuration.solve.opt_mode = "opt"
        c.configuration.solve.solve_limit = solve_limit
        c.configuration.solve.parallel_mode = int(config["number_cores"])
        c.configuration.solver.deletion = "basic,1"

        aset[3] = c

        c.add("prog", [], str(self.prog))

        def solver(c, om):
            c.ground([("prog", [])])
            self.grounded = True
            c.solve(on_model=om)

        t = threading.Thread(target=solver, args=(c, __on_model))
        t.start()
        t.join(timeout)
        self.timeout = t.is_alive()
        c.interrupt()
        t.join()

        aset[1] |= c.statistics["summary"]["models"]["optimal"] > 0
        aset[4] = c.statistics
        return aset



import queue
from aspmc.util import *
import networkx as nx
import numpy as np

class Node(object):    
    """A node class corresponding to nodes in a `ConstrainedSDD`.
    
    Args:
        type (:obj:`int`): Which type the node has. One of `Node.AND, Node.OR, Node.NEG, Node.IN`.
        name (:obj:`int`): Unique identifier of the node.
        children (:obj:`iterable`, optional): The children of the node. Defaults to `[]`.
        ancestors (:obj:`iterable`, optional): The ancestors of the node. Defaults to `[]`.

    Attributes:
        type (:obj:`int`): Which type the node has. One of `Node.AND, Node.OR, Node.NEG, Node.IN`.
        name (:obj:`int`): Unique identifier of the node. 
        If `type` is `Node.NEG` or `Node.IN` this is the name of the (negated) input variable.
        children (:obj:`iterable`): The children of the node.
        ancestors (:obj:`iterable`): The ancestors of the node. 
        vtree_node (:obj:`int`): The lowest vtree node that this node respects.

    """
    AND = 0
    """Node type `AND` with value 0. Means that the children are conjoined."""
    OR = 1
    """Node type `OR` with value 1. Means that the children are disjoined."""
    NEG = 2
    """Node type `NEG` with value 2. Means that this is a negative input variable."""
    IN = 3
    """Node type `IN` with value 3. Means that this is a positive input variable."""
    AND = 0
    OR = 1
    NEG = 2
    IN = 3
    def __init__(self, type, name, children = [], ancestors = []):
        self.type = type
        self.name = name
        self.vtree_node = None
        if self.type == Node.IN or self.type == Node.NEG:
            self.vtree_node = abs(name)
        self.children = children
        self.ancestors = ancestors

class ConstrainedSDD(object):    
    """A circuit class that can store SDDs. 

    Offers methods to perform algebraic model counting on the circuit in linear time in the size of the (smooth) circuit.
    Also offers the same methods statically to avoid storing the circuit in memory. 
    Then the algebraic model counting is performed while reading the circuit.

    Supports the d-DNNF format of c2d and miniC2D for constructing circuits.
    
    Args:
        path (:obj:`string`): The path to the file that contains the circuit.

    Attributes:
        literals (:obj:`list`): The literals in the circuit. The node for literal `v` is in `literals[2*(v-1)]`, the one for `-v` is in `literals[2*(v-1)+1]`
        nodes (:obj:`list`): All the nodes of the circuit that are not literals.
    """
    def __init__(self, path = None, v3 = None):
        if v3 != None:
            self.set_vtree(v3)
        if path != None:
            with open(path) as ddnnf:
                _, nr_nodes, nr_edges, nr_leafs = ddnnf.readline().split()
                self.literals = []
                for var in range(1, int(nr_leafs) + 1):
                    self.literals.append(Node(Node.IN, var))
                    self.literals.append(Node(Node.NEG, -var))
                self.nodes = []
                idx = 0
                for line in ddnnf:
                    line = line.strip().split()
                    if line[0] == 'L':
                        node = self.literals[2*(abs(int(line[1]))-1) + (1 if int(line[1]) < 0 else 0)]
                    elif line[0] == 'A':
                        node = Node(Node.AND, idx, children = [self.nodes[int(x)] for x in line[2:]])
                        for child in node.children:
                            child.ancestors.append(node)
                    elif line[0] == 'O':
                        node = Node(Node.OR, idx, children = [self.nodes[int(x)] for x in line[3:]])
                        for child in node.children:
                            child.ancestors.append(node)
                    self.nodes.append(node)
                    idx += 1

    def set_vtree(self, v3):
        self.vtree = v3
        self.vtree_graph = self.vtree.to_networkx()
        self.vtree.index()
        self.lca = [ [ -1 for _ in range(self.vtree.leaf_count()*2) ] for _ in range(self.vtree.leaf_count()*2) ]
        for p,l in nx.tree_all_pairs_lowest_common_ancestor(self.vtree_graph):
            self.lca[p[1]][p[0]] = l
            self.lca[p[0]][p[1]] = l


    def get_vtree_node(self, node):
        if node.vtree_node is None:
            node.vtree_node = self.lca[(node.children[0].vtree_node, node.children[1].vtree_node)]
        return node.vtree_node
        
    def separator_node(self, P):
        stack = queue.LifoQueue()
        cur = self.vtree
        down = True
        lefts = 0
        seen = set()
        while not (lefts == 0 and seen.issuperset(P)):
            if down:
                if cur.val is None:
                    stack.put(cur)
                    cur = cur.left
                    lefts += 1
                else:
                    seen.add(cur.val)
                    down = False
            else:
                last = cur
                cur = stack.get()
                if last == cur.left:
                    stack.put(cur)
                    cur = cur.right
                    lefts -= 1
                    down = True
        return cur.idx

    def wmc(self, weights, P, first_semiring, second_semiring, transform):        
        """Performs (two) algebraic model counting over the X/D-constrained SDD.
        
        Args:
            weights (:obj:`string`): The weights of the literals. The weight for literal `v` is in `weights[2*(v-1)]`, the one for `-v` is in `weights[2*(v-1)+1]`
            P (:obj:`iterable`): The set of variables that are quantified over the first semiring.
            first_semiring (:obj:`module`): The module of the first semiring.
            second_semiring (:obj:`module`): The module of the second semiring.
            transform (:obj:`string`): The transformation function used to transform a value from the second semiring into a value from the first semiring.
                Will be used as

                        f_transform = eval(transform)
                        transform = lambda x : first_semiring.from_value(f_transform(x))
        Returns:
            (:obj:`object`): The algebraic model count.
        """
        first_shape = (np.shape(weights[0])[0], ) + np.shape(first_semiring.one())
        second_shape = (np.shape(weights[0])[0], ) + np.shape(second_semiring.one())

        # initialize the transform function
        f_transform = eval(transform)
        transform = lambda x : first_semiring.from_value(f_transform(x))

        separator_node = self.separator_node(P)
        for i in range(len(self.literals)):
            if abs(to_dimacs(i)) in P:
                self.literals[i].weight = weights[i]
            elif self.lca[separator_node][abs(to_dimacs(i))] == separator_node: 
                self.literals[i].weight = weights[i]
            else:
                self.literals[i].weight = np.array([ transform(w) for w in weights[i] ])

        
        index_to_node = { node.idx : node for node in self.vtree.nodes() }
        index_to_node = [ index_to_node[i] for i in range(1, self.vtree.leaf_count()*2) ]
        def factor(i, j):
            if self.lca[i][j] == j: # i should always be above j
                i,j = j,i
            if self.lca[i][separator_node] == separator_node and i not in P:
                res = np.full(second_shape, second_semiring.one())
            else:
                res = np.full(first_shape, first_semiring.one())
            if i == j:
                return res
            upper = index_to_node[i - 1]
            lower = index_to_node[j - 1]
            res = np.full(first_shape, first_semiring.one())
            stack = queue.LifoQueue()
            cur = upper
            down = True
            while cur != upper or down:
                if down:
                    if cur == lower:
                        down = False
                    elif cur.val is None:
                        stack.put(cur)
                        cur = cur.left
                    else:
                        res *= self.literals[to_pos(cur.val)].weight + self.literals[neg(to_pos(cur.val))].weight
                        down = False
                else:
                    last = cur
                    cur = stack.get()
                    if cur.left == last:
                        stack.put(cur)
                        cur = cur.right
                        down = True
            return res

        todo = [ len(node.children) for node in self.nodes ]
        q = queue.Queue()
        for i in range(len(self.nodes)):
            if todo[i] == 0:
                q.put(self.nodes[i])
        while not q.empty():
            node = q.get()
            vtree_node = self.get_vtree_node(node)
            for anc in node.ancestors:
                todo[anc.name] -= 1
                if todo[anc.name] == 0:
                    q.put(anc)
            if self.lca[separator_node][vtree_node] != separator_node:
                if node.type == Node.AND:
                    node.weight = np.empty(first_shape, dtype=first_semiring.dtype)
                    node.weight[:] = first_semiring.one()
                    for child in node.children:
                        child_vtree_node = self.get_vtree_node(child)
                        if self.lca[child_vtree_node][separator_node] == separator_node and child_vtree_node not in P:
                            node.weight *= np.array([ transform(w) for w in child.weight ])
                        else:
                            node.weight *= child.weight
                elif node.type == Node.OR:
                    node.weight = np.empty(first_shape, dtype=first_semiring.dtype)
                    node.weight[:] = first_semiring.zero()
                    for child in node.children:
                        node.weight += child.weight*factor(vtree_node, child.vtree_node)
            else:                
                if node.type == Node.AND:
                    node.weight = np.empty(second_shape, dtype=second_semiring.dtype)
                    node.weight[:] = second_semiring.one()
                    for child in node.children:
                        node.weight *= child.weight
                elif node.type == Node.OR:
                    node.weight = np.empty(second_shape, dtype=second_semiring.dtype)
                    node.weight[:] = second_semiring.zero()
                    for child in node.children:
                        node.weight += child.weight*factor(vtree_node, child.vtree_node)
        node.weight *= factor(node.vtree_node, self.vtree.idx)
        return node.weight

    def parse_wmc(self, path, weights, cnf): # path, weights, P, first_semiring, second_semiring, transform):
        """Performs (two) algebraic model counting over an X/D-constrained SDD while parsing it.
        
        Args:
            path (:obj:`string`): The path to the file that contains the SDD.
            weights (:obj:`string`): The weights of the literals. The weight for literal `v` is in `weights[2*(v-1)]`, the one for `-v` is in `weights[2*(v-1)+1]`
            P (:obj:`iterable`): The set of variables that are quantified over the first semiring.
            first_semiring (:obj:`module`): The module of the first semiring.
            second_semiring (:obj:`module`): The module of the second semiring.
            transform (:obj:`string`): The transformation function used to transform a value from the second semiring into a value from the first semiring.
                Will be used as

                        f_transform = eval(transform)
                        transform = lambda x : first_semiring.from_value(f_transform(x))
        Returns:
            (:obj:`object`): The algebraic model count.
        """
        shapes = [ (np.shape(weights[0])[0], ) + np.shape(semiring.one()) for semiring in cnf.semirings ]

        first_semiring = cnf.semirings[0]
        second_semiring = cnf.semirings[1]
        
        transform = cnf.transforms[0]

        P = cnf.quantified[0]        
        # initialize the transform function
        f_transform = eval(transform)
        transform = lambda x : first_semiring.from_value(f_transform(x))

        separator_node = self.separator_node(P)
        for i in range(len(weights)):
            if abs(to_dimacs(i)) not in P and self.lca[separator_node][abs(to_dimacs(i))] != separator_node: 
                weights[i] = np.array([ transform(w) for w in weights[i] ])
        
        index_to_node = { node.idx : node for node in self.vtree }
        index_to_node = [ index_to_node[i] for i in range(1, self.vtree.leaf_count()*2) ]

        def factor(i, j):
            if self.lca[i][j] == j: # i should always be above j
                i,j = j,i
            if self.lca[i][separator_node] == separator_node and i not in P:
                res = np.full(shapes[1], second_semiring.one())
            else:
                res = np.full(shapes[0], first_semiring.one())
            if i == j:
                return res
            upper = index_to_node[i - 1]
            lower = index_to_node[j - 1]
            stack = queue.LifoQueue()
            cur = upper
            down = True
            while cur != upper or down:
                if down:
                    if cur == lower:
                        down = False
                    elif cur.val is None:
                        stack.put(cur)
                        cur = cur.left
                    else:
                        res *= weights[to_pos(cur.val)] + weights[neg(to_pos(cur.val))]
                        down = False
                else:
                    last = cur
                    cur = stack.get()
                    if cur.left == last:
                        stack.put(cur)
                        cur = cur.right
                        down = True
            return res

        with open(path) as ddnnf:
            _, nr_nodes, nr_edges, nr_leafs = ddnnf.readline().split()
            mem = []
            vtree_nodes = []
            idx = 0
            for line in ddnnf:
                line = line.strip().split()
                if line[0] == 'L':
                    val = weights[to_pos(int(line[1]))]
                    vtree_node = abs(int(line[1]))
                else:
                    if line[0] == 'A':
                        vtree_node = self.lca[vtree_nodes[int(line[2])]][vtree_nodes[int(line[3])]]
                        if self.lca[separator_node][vtree_node] != separator_node:
                            val = np.empty(shapes[0], dtype=first_semiring.dtype)
                            val[:] = first_semiring.one()
                            for x in line[2:]:
                                x = int(x)
                                child_vtree_node = vtree_nodes[x]
                                if self.lca[child_vtree_node][separator_node] == separator_node and child_vtree_node not in P:
                                    val *= np.array([ transform(w) for w in mem[x] ])
                                else:
                                    val *= mem[x]
                        else:
                            val = np.empty(shapes[1], dtype=second_semiring.dtype)
                            val[:] = second_semiring.one()
                            for x in line[2:]:
                                val *= mem[int(x)]
                    elif line[0] == 'O':
                        vtree_node = self.lca[vtree_nodes[int(line[3])]][vtree_nodes[int(line[4])]]
                        if self.lca[separator_node][vtree_node] != separator_node:
                            val = np.empty(shapes[0], dtype=first_semiring.dtype)
                            val[:] = first_semiring.zero()
                            for x in line[3:]:
                                val += mem[int(x)]*factor(vtree_node, vtree_nodes[int(x)])
                        else:                
                            val = np.empty(shapes[1], dtype=second_semiring.dtype)
                            val[:] = second_semiring.zero()
                            for x in line[3:]:
                                val += mem[int(x)]*factor(vtree_node, vtree_nodes[int(x)])
                mem.append(val)
                vtree_nodes.append(vtree_node)
                idx += 1
            mem[idx - 1] *= factor(vtree_node, self.vtree.idx)
            return mem[idx - 1]




"""The module aspmc.graph contains all the graph related classes:

    * a hypergraph class in aspmc.graph.hypergraph, which can be used to generated tree decompositions.
    * a binary tree class in aspmc.graph.bintree, which is used for D/Vtrees.
    * a tree decomposition class in aspmc.graph.treedecomposition, which can generated tree decompositions from different inputs.
        Also it offers many different utilities.  
"""

from . import *


"""
Program module providing the algebraic progam class.
"""

import re
import numpy as np
import logging
import math

from aspmc.parsing.clingoparser.clingoext import ClingoRule, Control

from aspmc.programs.algebraicprogram import AlgebraicProgram

import aspmc.semirings.probabilistic as probabilistic
import aspmc.semirings.maxtimes as maxtimes


from aspmc.util import *

logger = logging.getLogger("aspmc")

class MPEProblogProgram(AlgebraicProgram):
    """A class for programs with weights over semirings. 

    Should be specified in ProbLog syntax, but allows negations and negative cycles.

    Annotated disjunctions are theoretically supported over any semiring but the results are likely to 
    be wrong over semirings that differ from the probabilistic semiring.

    Grounding of these programs (and subclasses thereof) should follow the following strategy:

    * `_prepare_grounding(self, program)` should take the output of the parser 
        (i.e. a list of rules and special objects) and process all the rules and special objects
        transforming them either into other rules or into strings that can be given to the grounder.
    * the output of `_prepare_grounding(self, program)` is transformed to one program string via

            '\\n'.join([ str(r) for r in program ])
        
        This string will be given to the grounder, which produces a clingo control object.
    * `_process_grounding(self, clingo_control)` should take this clingo control object and process the
        grounding in an appropriate way (and draw some information from it optionally about weights, special objects).
        The resulting processed clingo_control object must only know about the 
        rules that should be seen in the base program class.

    Thus, subclasses can override `_prepare_grounding` and `_process_grounding` (and optionally call the superclass methods) 
    to handle their extras. See aspmc.programs.meuprogram or aspmc.programs.smprogram for examples.

    Args:
        program_str (:obj:`string`): A string containing a part of the program in ProbLog syntax. 
        May be the empty string.
        program_files (:obj:`list`): A list of string that are paths to files which contain programs in 
        ProbLog syntax that should be included. May be an empty list.
        semiring (:obj:`module`): The semiring module to be used. See aspmc.semiring for how they should look.

    Attributes:
        semiring (:obj:`module`): The semiring module to be used. 
        weights (:obj:`dict`): The dictionary from atom names to their weight.
        queries (:obj:`list`): The list of atoms to be queries in their string representation.
    """
    def __init__(self, program_str, program_files):
        self.evidence = {}
        AlgebraicProgram.__init__(self, program_str, program_files, probabilistic)
        self.semiring = maxtimes
        self.original_weights = self.weights
        self.weights = { name : maxtimes.MaxTimesFloat(weight) for name, weight in self.weights.items() }

    def _prepare_grounding(self, program):
        new_program = []
        for r in program:
            if isinstance(r, str):
                new_program.append(r)
            elif r.head is not None and r.head[0].predicate == "evidence":
                atom = r.head[0].inputs[0]
                new_program.append(f":-{'' if atom.negated else 'not'} {str(atom)[4:] if atom.negated else str(atom)}.")
                phase = atom.negated
                atom.negated = False
                self.evidence[str(atom)] = phase
            else:
                new_program.append(r)
        return super()._prepare_grounding(new_program)

    def _prog_string(self, program):
        program = [ r for r in program if len(r.head) > 0 ]
        result = AlgebraicProgram._prog_string(self, program)
        for name in self.evidence:
            result += "evidence("
            if self.evidence[name]:
                result += "\\+"
            result += f"{name}).\n"
        return result

    def _finalize_cnf(self):
        weight_list = self.get_weights()
        for v in range(self._cnf.nr_vars*2):
            self._cnf.weights[to_dimacs(v)] = weight_list[v]
        self._cnf.semirings = [ self.semiring ]
        self._cnf.quantified = [ list(range(1, self._cnf.nr_vars + 1)) ]
        query_cnt = max(len(self.queries), 1)
        for v in self._deriv:
            if all(self._cnf.weights[v][i] == self.semiring.one() and self._cnf.weights[-v][i] == self.semiring.one() for i in range(query_cnt)):
                self._cnf.auxilliary.add(v)

    def get_weights(self):
        query_cnt = 1
        varMap = { name : var for var, name in self._nameMap.items() }
        weight_list = [ np.full(query_cnt, self.semiring.one(), dtype=self.semiring.dtype) for _ in range(self._cnf.nr_vars*2) ]
        for (name, phase) in self.weights:
            if phase:
                weight_list[to_pos(varMap[name])] = np.full(query_cnt, self.weights[(name, phase)], dtype=self.semiring.dtype)
            else:
                weight_list[neg(to_pos(varMap[name]))] = np.full(query_cnt, self.weights[(name, phase)], dtype=self.semiring.dtype)
        return weight_list

    def get_queries(self):
        return self.queries



names = []

class MaxTimesFloat(object):
    def __init__(self, value, decisions):
        self.value = value
        self.decisions = decisions

    def __add__(self, other):
        return self if self.value >= other.value else other

    def __iadd__(self, other):
        if self.value < other.value:
            self.value = other.value
            self.decisions = other.decisions
        return self

    def __mul__(self, other):
        return MaxTimesFloat(self.value * other.value, self.decisions | other.decisions)

    def __imul__(self, other):
        self.value *= other.value
        self.decisions |= other.decisions
        return self

    def __str__(self):
        decisions = [ names[i] for i in range(len(names)) if self.decisions & 2**i ]
        decisions = ", ".join(decisions)
        return f"{self.value} with true atoms: {decisions}"

    def __repr__(self):
        return str(self)

def parse(value, atom = None):
    value = value[1:-1]
    value = value.split(',')
    return MaxTimesFloat(float(value[0]), int(value[1]))

def from_value(value):
    return MaxTimesFloat(value, 0)

def negate(value):
    return one()

def to_string(value):
    return f"({value.value},{value.decisions})"

def zero():
    return MaxTimesFloat(float("-inf"), 0)
def one():
    return MaxTimesFloat(1, 0)

dtype = object
pattern = '([+-]?([0-9]*[.])?[0-9]+,[0-9]+)'


"""
Program module providing the optimizing progam class.
"""

import re
import numpy as np
import logging

from aspmc.programs.program import Program, UnsupportedException

from aspmc.parsing.clingoparser.clingoext import ClingoRule, Control

from lark import Lark
from aspmc.parsing.lark_optparser import GRAMMAR, OptProgramTransformer, WeakConstraint
import aspmc.programs.grounder as grounder

import aspmc.semirings.minplus as semiring

from aspmc.util import *
from aspmc.programs.naming import *

logger = logging.getLogger("aspmc")

class OptProgram(Program):
    """A class for programs that state optimization problems. 

    Should be specified using only normal rules, choice rules and weak constraints.

    Grounding of these programs (and subclasses thereof) should follow the following strategy:

    * `_prepare_grounding(self, program)` should take the output of the parser 
        (i.e. a list of rules and special objects) and process all the rules and special objects
        transforming them either into other rules or into strings that can be given to the grounder.
    * the output of `_prepare_grounding(self, program)` is transformed to one program string via

            '\\n'.join([ str(r) for r in program ])
        
        This string will be given to the grounder, which produces a clingo control object.
    * `_process_grounding(self, clingo_control)` should take this clingo control object and process the
        grounding in an appropriate way (and draw some information from it optionally about weights, special objects).
        The resulting processed clingo_control object must only know about the 
        rules that should be seen in the base program class.

    Thus, subclasses can override `_prepare_grounding` and `_process_grounding` (and optionally call the superclass methods) 
    to handle their extras. See aspmc.programs.meuprogram or aspmc.programs.smprogram for examples.

    Args:
        program_str (:obj:`string`): A string containing a part of the program. 
        May be the empty string.
        program_files (:obj:`list`): A list of string that are paths to files which contain programs. May be an empty list.

    Attributes:
        semiring (:obj:`module`): The semiring module to be used. 
        weights (:obj:`dict`): The dictionary from atom names to their weight.
    """
    def __init__(self, program_str, program_files):
        self.semiring = semiring
        self.weights = {}
        for path in program_files:
            with open(path) as file_:
                program_str += file_.read()
        # parse
        my_grammar = GRAMMAR
        parser = Lark(my_grammar, start='program', parser='lalr', transformer=OptProgramTransformer())
        program = parser.parse(program_str)

        # ground
        clingo_control = Control()
        self._ground(clingo_control, program)

        # initialize the superclass
        Program.__init__(self, clingo_control = clingo_control)

    def _ground(self, clingo_control, program):
        # do the grounding in three steps:
        # 1. transform the parsed rules into asp rules
        program = self._prepare_grounding(program)
        # 2. give the asp rules to the grounder
        clingo_str = '\n'.join([ str(r) for r in program ])
        grounder.ground(clingo_control, program_str = clingo_str, program_files = [])
        # 3. take care of possible extras
        self._process_grounding(clingo_control)

    def _prepare_grounding(self, program):
        new_program = []
        for r in program:
            if isinstance(r, str):
                new_program.append(r)
            elif isinstance(r, WeakConstraint):
                new_program.append(f"{INTERNAL_PENALTY}({','.join([ str(term) for term in r.terms ])},{r.weight}):-{','.join([ str(atom) for atom in r.body ])}.")
            else:
                new_program.append(r)
        return new_program

    def _process_grounding(self, clingo_control):
        symbol_map = {}
        for sym in clingo_control.symbolic_atoms:
            symbol_map[sym.literal] = str(sym.symbol)
        for o in clingo_control.ground_program.objects:
            if isinstance(o, ClingoRule):
                head_name = "" if len(o.head) == 0 else symbol_map[abs(o.head[0])]
                # check if the rule corresponds to a weak constraint
                if head_name.startswith(INTERNAL_PENALTY):
                    # find out the penalty of this atom
                    idx = head_name.rfind(",")
                    penalty = semiring.MinPlusFloat(float(head_name[idx+1:-1]))
                    self.weights[head_name] = penalty
                

    def _prog_string(self, program):
        # TODO: return the weak constraint program again
        result = ""
        for v in self._guess:
            result += f"{{{self._external_name(v)}}}.\n"
        for r in program:
            result += ";".join([self._external_name(v) for v in r.head])
            if len(r.body) > 0:
                result += ":-"
                result += ",".join([("not " if v < 0 else "") + self._external_name(abs(v)) for v in r.body])
            result += ".\n"
        return result

    def _finalize_cnf(self):
        weight_list = self.get_weights()
        for v in range(self._cnf.nr_vars*2):
            self._cnf.weights[to_dimacs(v)] = weight_list[v]
        self._cnf.semirings = [ self.semiring ]
        self._cnf.quantified = [ list(range(1, self._cnf.nr_vars + 1)) ]
        query_cnt = 1
        for v in self._deriv:
            if all(self._cnf.weights[v][i] == self.semiring.one() and self._cnf.weights[-v][i] == self.semiring.one() for i in range(query_cnt)):
                self._cnf.auxilliary.add(v)

    def get_weights(self):
        query_cnt = 1
        varMap = { name : var for var, name in self._nameMap.items() }
        weight_list = [ np.full(query_cnt, self.semiring.one(), dtype=self.semiring.dtype) for _ in range(self._cnf.nr_vars*2) ]
        for name in self.weights:
            weight_list[to_pos(varMap[name])] = np.full(query_cnt, self.weights[name], dtype=self.semiring.dtype)
        return weight_list

    def get_queries(self):
        return []

    # def solve_clingo(self):
    #     import clingo
    #     control = clingo.Control()
    #     control.add("base", [], self._prog_string(self._program))
    #     control.ground([('base', [])])



def ground(control, program_str = None, program_files = []):
    if program_str != None:
        control.add("base", [], program_str)
    for path in program_files:
        with open(path) as file_:
            control.add("base", [], file_.read())
    control.ground([('base', [])])
    return control





import numpy as np

def parse(value, atom = None):
    value = value[1:-1]
    value = value.split(',')
    return np.array([float(value[0]), float(value[1])])

def from_value(value):
    return np.array(value)

def negate(value):
    return value.copy()

def to_string(value):
    return f"({value[0]},{value[1]})"

def is_idempotent():
    return False

def zero():
    return np.array([0.0, 0.0])
def one():
    return np.array([1.0, 1.0])
    
dtype = float
pattern = '[(][0-9]+[,][0-9]+[)]'
